from osiris_skretrieval.osiris import Osiris
from osiris_skretrieval.osirismodel import OSIRISModel, IdealOSIRISModel
import sasktran as sk
from osiris_skretrieval.osiris_no2 import OSIRISNO2Retrieval
import numpy as np
from skretrieval.retrieval.rodgers import Rodgers
import xarray as xr
from osiris_skretrieval.osiris_spectral_fit import spectral_fit
from copy import copy
from osirisl1services.readlevel1 import open_level1_ecmwf
import julian


def load_osiris(scan_no, fwhm_type, apply_rsas, load_invalid_osiris_data):
    if fwhm_type == 'l1_fit':
        load_psf = True
    else:
        load_psf = False

    os_data = Osiris(scan_no, load_psf=load_psf, apply_rsas=apply_rsas, load_invalid=load_invalid_osiris_data)

    if fwhm_type == 'spectral_fit':
        spectral_fits = spectral_fit(os_data.skretrieval_l1(), fit_alt_range=(40000, 50000))
        os_data.spectrograph_data['fwhm'].values = np.ones_like(os_data.spectrograph_data['fwhm'].values) * spectral_fits['fwhm']

    return os_data


def retrieve_no2_scan(scan_no, fwhm_type='spectral_fit', apply_rsas=True, tikh_factor=100,
                      include_polarization_fit=False, retrieve_on_osiris_grid=False,
                      include_ozone=True,
                      max_iter=10,
                      use_ideal_osiris=True,
                      upper_altitude_bound=48500,
                      high_alt_normalization_range=5000,
                      initial_lm_damping=0.5,
                      good_pixel_fraction=0.9999,
                      load_invalid_osiris_data=False,
                      osiris_aerosol_folder=None,
                      wavel_range=(434.7, 476.5)):

    os_data = load_osiris(scan_no=scan_no, fwhm_type=fwhm_type, apply_rsas=apply_rsas,
                          load_invalid_osiris_data=load_invalid_osiris_data)

    os_wavel = os_data.skretrieval_l1().data['wavelength'].values
    os_wavel = os_wavel[(os_wavel > wavel_range[0]) & (os_wavel < wavel_range[1])]

    skl1 = os_data.skretrieval_l1()
    tl = os_data.skretrieval_l1().tangent_locations()

    mean_mjd = float(skl1.data.where(tl.altitude < 40000).mjd.mean())
    mean_lat = float(tl.where(tl.altitude < 40000).latitude.mean())
    mean_lon = float(tl.where(tl.altitude < 40000).longitude.mean())

    pratmo_sim = sk.Pratmo()

    no2_alts = np.linspace(500, 99500, 100)
    no2_vals = pratmo_sim.get_parameter('SKCLIMATOLOGY_NO2_CM3', mean_lat, mean_lon, no2_alts, mean_mjd)

    convolved_no2 = sk.OpticalPropertyConvolved(sk.NO2Vandaele1998(), os_data.skretrieval_l1().data['wavelength'].values, os_data.skretrieval_l1().data.fwhm.values / 2.355)
    convolved_o3 = sk.OpticalPropertyConvolved(sk.O3DBM(), os_data.skretrieval_l1().data['wavelength'].values, os_data.skretrieval_l1().data.fwhm.values / 2.355)

    atmo_ret = sk.Atmosphere()

    no2_user_def = sk.ClimatologyUserDefined(no2_alts, {'SKCLIMATOLOGY_NO2_CM3': no2_vals})

    ecmwf_data = open_level1_ecmwf(scan_no)
    pressure = ecmwf_data.density / 100 * ecmwf_data.temperature * 8.31446261815324 / 6.02214086e23 * 1e6

    background = sk.ClimatologyUserDefined(ecmwf_data.Altitude.values,
                                           {'SKCLIMATOLOGY_PRESSURE_PA': pressure.values * 100,
                                            'SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3': ecmwf_data.density.values,
                                            'SKCLIMATOLOGY_TEMPERATURE_K': ecmwf_data.temperature.values})

    atmo_ret['air'] = sk.Species(sk.Rayleigh(), background, species='SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3')
    atmo_ret['no2'] = sk.Species(convolved_no2, no2_user_def)

    if include_ozone:
        atmo_ret['o3'] = sk.Species(convolved_o3, sk.Labow())

    if osiris_aerosol_folder is not None:
        date = julian.from_jd(mean_mjd, fmt='mjd')
        aerosol_data = xr.open_mfdataset(osiris_aerosol_folder + '/AEROSOL-L2-LP-OSIRIS_ODIN-SASK_v7_0-{}{}.nc'.format(str(date.year).zfill(4), str(date.month).zfill(2)), combine='by_coords', concat_dim='profile_id')

        aerosol_data = aerosol_data.sel(profile_id=scan_no)

        particle_size_dist = sk.ClimatologyUserDefined(altitudes=[0, 100000],
                                                       values={'SKCLIMATOLOGY_LOGNORMAL_MODEWIDTH': [1.6, 1.6],
                                                               'SKCLIMATOLOGY_LOGNORMAL_MODERADIUS_MICRONS': [0.08,
                                                                                                              0.08]})

        aero_opt_prop = sk.MieAerosol(particle_size_dist, 'H2SO4')
        xs_aero = aero_opt_prop.calculate_cross_sections(sk.MSIS90(), 0, 0, 20000, 54372, 750).total

        numden_aero = aerosol_data._extinction_rtm / xs_aero / 1e5

        rtm_alts = np.arange(0, 100000, 500)
        rtm_aero = np.interp(rtm_alts, aerosol_data.altitude.values * 1000, numden_aero.values, left=numden_aero[0], right=0)

        aero_clim = sk.ClimatologyUserDefined(rtm_alts, {'aerosol': rtm_aero})

        atmo_ret['aerosol'] = sk.Species(aero_opt_prop, aero_clim)


    atmo_ret.atmospheric_state = background

    no2_ret = OSIRISNO2Retrieval(atmo_ret['no2'], os_data.skretrieval_l1(),
                                 tikh_factor=tikh_factor, high_alt_normalize=True,
                                 no2_opticalproperty=convolved_no2,
                                 o3_opticalproperty=convolved_o3,
                                 include_polarization_fit=include_polarization_fit,
                                 retrieve_on_osiris_grid=retrieve_on_osiris_grid,
                                 v6_normalization_altitude=False,
                                 high_alt_normalization_range=high_alt_normalization_range,
                                 upper_altitude_bound=upper_altitude_bound,
                                 good_pixel_fraction=good_pixel_fraction)

    options=dict()
    options['calcwf'] = 2
    options['wfheights'] = no2_ret._retrieval_altitudes
    options['wfwidths'] = np.gradient(no2_ret._retrieval_altitudes)
    options['numordersofscatter'] = 50

    atmo_ret.wf_species = 'no2'

    if use_ideal_osiris:
        os_model_ret = IdealOSIRISModel(os_data, atmo_ret, options, os_wavel)
    else:
        os_model_ret = OSIRISModel(os_data, atmo_ret, model_wavel=os_wavel, meas_wavel=os_wavel, options=options,
                                   model_vertical_resolution_rad=0.0002)

    ret = Rodgers(max_iter=max_iter, lm_damping=initial_lm_damping, iterative_update_lm=True, convergence_factor=1.01)
    ret_results = ret.retrieve(os_data.skretrieval_l1(), os_model_ret, no2_ret)

    ret_no2 = no2_user_def.get_parameter('SKCLIMATOLOGY_NO2_CM3', 0, 0, no2_alts, 54372)
    valid_no2 = copy(ret_no2)

    y_meas = no2_ret.measurement_vector(os_data.skretrieval_l1())
    low_alt = y_meas['y_altitudes'][~np.isnan(y_meas['y'])][0]
    high_alt = 35000

    valid_no2[no2_alts < low_alt] = np.nan
    valid_no2[no2_alts > high_alt] = np.nan

    error_covar = no2_ret.map_error_covariance(ret_results['error_covariance_from_noise'])

    pressure_on_grid = np.interp(no2_alts, ecmwf_data.Altitude.values, pressure.values)
    numden_on_grid = np.interp(no2_alts, ecmwf_data.Altitude.values, ecmwf_data.density.values)
    temp_on_grid = np.interp(no2_alts, ecmwf_data.Altitude.values, ecmwf_data.temperature.values)

    ret_data = xr.Dataset({'no2': (['altitude'], ret_no2),
                           'temperature': (['altitude'], temp_on_grid),
                           'pressure': (['altitude'], pressure_on_grid),
                           'background_density': (['altitude'], numden_on_grid),
                           'valid_no2': (['altitude'], valid_no2),
                           'chi_sq': ret_results['chi_sq'],
                           'chi_sq_linear': ret_results['chi_sq_linear'],
                           'chi_sq_meas': ret_results['chi_sq_meas'],
                           'chi_sq_meas_linear': ret_results['chi_sq_meas_linear'],
                           'scanno': scan_no,
                           'mean_mjd': mean_mjd,
                           'mean_lat': mean_lat,
                           'mean_lon': mean_lon,
                           'error_covariance': (['altitude', 'altitude_2'], error_covar)},
                          coords={'altitude': no2_alts,
                                  'altitude_2': no2_alts})

    diag_data = xr.Dataset({'averaging_kernel': (['ret_altitudes', 'true_altitudes'], ret_results['averaging_kernel']),
                            'y_ret': (['y_altitude'], ret_results['ys'][-1]),
                            'y_meas': (['y_altitude'], ret_results['y_meas']),
                            'fit_residuals': (['y_altitude', 'fit_wavelength'], y_meas['fit_residuals']),
                            'good_pixel_fraction': (['y_altitude'], y_meas['good_pixel_fraction'])},
                          coords={'y_altitude': y_meas['y_altitudes'],
                                  'fit_wavelength': y_meas['fit_residual_wavelengths'],
                                  'ret_altitudes': no2_ret._retrieval_altitudes,
                                  'true_altitudes': no2_ret._retrieval_altitudes})

    return ret_data, diag_data